@@ -8,7 +8,7 @@ use serde::{Serialize, Deserialize};
 
 use std::cmp::Ordering;
 
-use parking_lot::{RwLock, Mutex};
+use parking_lot::{RwLock, Mutex, RwLockReadGuard};
 use std::sync::Arc;
 use rayon::prelude::*;
 use std::sync::mpsc::channel;
@@ -323,6 +323,9 @@ impl LayerGenerator {
 
 // ====================================================================
 
+/// A short-hand for points in a layer
+type Layer<T> = Vec<Arc<Point<T>>>;
+
 /// a structure for indexation of points in layer
 #[allow(unused)]
 pub struct PointIndexation<T:Clone+Send+Sync> {
@@ -331,7 +334,7 @@ pub struct PointIndexation<T:Clone+Send+Sync> {
     ///
     pub(crate) max_layer: usize,
     /// needs at least one representation of points. points_by_layers\[i\] gives the points in layer i
-    pub(crate) points_by_layer: Arc<RwLock<Vec<Vec<Arc<Point<T>> >> >>,
+    pub(crate) points_by_layer: Arc<RwLock<Vec< Layer<T> > >>,
     /// utility to generate a level
     pub(crate) layer_g: LayerGenerator,
     /// number of points in indexed structure
@@ -441,6 +444,18 @@ impl<T:Clone+Send+Sync> PointIndexation<T> {
         *self.nb_point.read()
     }
 
+    /// returns the number of points in a given layer, 0 on a bad layer num
+    pub fn get_layer_nb_point(&self, layer : usize) -> usize {
+        let nb_layer = self.points_by_layer.read().len();
+        if layer < nb_layer {
+            self.points_by_layer.read()[layer].len()
+        }
+        else {
+            0
+        }
+    } // end of get_layer_nb_point
+
+
     /// returns the size of data vector in graph if any, else return 0
     pub fn get_data_dimension(&self) -> usize {
         let ep = self.entry_point.read();
@@ -450,6 +465,11 @@ impl<T:Clone+Send+Sync> PointIndexation<T> {
             };
         dim
     }
+
+    /// get an iterator on the points stored in a given layer
+    pub fn get_layer_iterator(&self, layer : usize) -> IterPointLayer<T> {
+        IterPointLayer::new(&self, layer)
+    }  // end of get_layer_iterator
 }  // end of impl PointIndexation
 
 
@@ -458,16 +478,18 @@ impl<T:Clone+Send+Sync> PointIndexation<T> {
 
 /// an iterator on points stored.
 /// The iteration begins at level 0 (most populated level) and goes upward in levels.
-/// Must not be used during parallel insertion.
+/// The iterator takes a ReadGuard on the PointIndexation structure
 pub struct IterPoint<'a,T:Clone+Send+Sync> {
     point_indexation : &'a PointIndexation<T>,
+    pi_guard :  RwLockReadGuard<'a, Vec<Layer<T>> >,
     layer:i64,
     slot_in_layer:i64,
 }
 
 impl <'a, T:Clone+Send+Sync> IterPoint<'a,T>{
     pub fn new(point_indexation : &'a PointIndexation<T>) -> Self {
-        IterPoint{ point_indexation, layer:-1, slot_in_layer : -1 }
+        let pi_guard : RwLockReadGuard<Vec<Layer<T>> > = point_indexation.points_by_layer.read();
+        IterPoint{ point_indexation, pi_guard, layer:-1, slot_in_layer : -1 }
     }
 }  // end of block impl IterPoint
 
@@ -481,10 +503,10 @@ impl <'a,T:Clone+Send+Sync> Iterator for IterPoint<'a,T> {
             self.layer = 0;
             self.slot_in_layer = 0;         
         }
-        if (self.slot_in_layer  as usize) < self.point_indexation.points_by_layer.read()[self.layer as usize].len() {
+        if (self.slot_in_layer  as usize) < self.pi_guard[self.layer as usize].len() {
             let slot = self.slot_in_layer as usize;
             self.slot_in_layer += 1;
-            return Some(self.point_indexation.points_by_layer.read()[self.layer as usize][slot].clone());
+            return Some(self.pi_guard[self.layer as usize][slot].clone());
         }
         else {
             self.slot_in_layer = 0;
@@ -521,8 +543,45 @@ impl<'a,T:Clone+Send+Sync> IntoIterator for &'a PointIndexation<T> {
         IterPoint::new(self)
     }
 
+}  // end of IntoIterator for &'a PointIndexation<T> 
+
+
+
+/// An iterator on points stored in a given layer
+/// The iterator stores a ReadGuard on the structure PointIndexation
+pub struct IterPointLayer<'a,T:Clone+Send+Sync> {
+    _point_indexation : &'a PointIndexation<T>,
+    pi_guard : RwLockReadGuard<'a, Vec<Layer<T>> >,
+    layer:usize,
+    slot_in_layer:usize,
 }
 
+impl <'a, T:Clone+Send+Sync> IterPointLayer<'a,T>{
+    pub fn new(point_indexation : &'a PointIndexation<T>, layer : usize) -> Self {
+        let pi_guard : RwLockReadGuard<Vec<Layer<T>> > = point_indexation.points_by_layer.read();
+        IterPointLayer{ _point_indexation : point_indexation, pi_guard, layer:layer, slot_in_layer : 0 }
+    }
+}  // end of block impl IterPointLayer
+
+
+/// iterator for layer 0 to upper layer.
+impl <'a,T:Clone+Send+Sync> Iterator for IterPointLayer<'a,T> {
+    type Item = Arc<Point<T>>;
+    //
+    fn next(&mut self) -> Option<Self::Item> {
+        if (self.slot_in_layer  as usize) < self.pi_guard[self.layer as usize].len() {
+            let slot = self.slot_in_layer as usize;
+            self.slot_in_layer += 1;
+            return Some(self.pi_guard[self.layer as usize][slot].clone());
+        }
+        else {
+            return None;
+        }
+    }  // end of next
+
+} // end of impl Iterator
+
+
 
 // ============================================================================================
 
@@ -652,10 +711,19 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
         }
     }  // end of set_scale_modification
 
+
     // here we could pass a point_id_with_order instead of entry_point_id: PointId
     // The efficacity depends on greedy part depends on how near entry point is from point.
     // ef is the number of points to return
     // The method returns a BinaryHeap with positive distances. The caller must transforms it according its need
+    //** NOTE: the entry point is pushed into returned point at the beginning of the function, but in fact entry_point is in a layer
+    //** with higher (one more) index than the argument layer. If the greedy search matches a sufficiently large number of points
+    //** nearer to point searched (arg point) than entry_point it will finally get popped up from the heap of returned points
+    //** but otherwise it will stay in the binary heap and so we can have a point in neighbours that is in fact in a layer
+    //** above the one we search in. 
+    //** The guarantee is that the binary heap will return points in layer 
+    //** with a larger index, although we can expect that most often (at least in densely populated layers) the returned
+    //** points will be found in searched layer 
     ///
     /// Greedy algorithm n° 2 in Malkov paper.
     /// search in a layer (layer) for the ef points nearest a point to be inserted in hnsw.
@@ -663,7 +731,7 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
         //
         trace!("entering search_layer with entry_point_id {:?} layer : {:?} ef {:?} ", entry_point.p_id, layer, ef);
         //
-        // here we allocate a skiplist on values not on reference beccause we want to return
+        // here we allocate a binary_heap on values not on reference beccause we want to return
         // log2(skiplist_size) must be greater than 1.
         let skiplist_size = ef.max(2);
         // we will store positive distances in this one
@@ -686,14 +754,13 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
         visited_point_id.insert(entry_point.p_id, Arc::clone(&entry_point));
         //
         let mut candidate_points = BinaryHeap::<Arc<PointWithOrder<T>> >::with_capacity(skiplist_size);
-
         candidate_points.push(Arc::new(PointWithOrder::new(&entry_point, -dist_to_entry_point)));
         return_points.push(Arc::new(PointWithOrder::new(&entry_point, dist_to_entry_point)));
         // at the beginning candidate_points contains point passed as arg in layer entry_point_id.0
         while candidate_points.len() > 0 {
             // get nearest point in candidate_points
             let c = candidate_points.pop().unwrap();
-           // f farthest point to 
+            // f farthest point to 
             let f = return_points.peek().unwrap();
             assert!(f.dist_to_ref >= 0.);
             assert!(c.dist_to_ref <= 0.);
@@ -788,6 +855,8 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
             // sorted_points = from_positive_binaryheap_to_negative_binary_heap(&mut sorted_points);
             // 
             if let Some(ep) = sorted_points.pop() {
+                // useful for projecting lower layer to upper layer. keep track of points encountered.
+                new_point.neighbours.write()[l as usize].push(Arc::clone(&ep));
                 // get the lowest distance point
                 let tmp_dist = self.dist_f.eval(data , & ep.point_ref.v);
                 if tmp_dist < dist_to_entry {
@@ -837,7 +906,7 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
         // new_point has been inserted at the beginning in table
         // so that we can call reverse_update_neighborhoodwe consitently
         // now reverse update of neighbours.
-        self.reverse_update_neighborhood_simple (Arc::clone(&new_point));
+        self.reverse_update_neighborhood_simple(Arc::clone(&new_point));
         //
         self.layer_indexed_points.check_entry_point(&new_point);
         //
@@ -867,7 +936,8 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
                     let q_point = &q.point_ref;
                     let mut q_point_neighbours = q_point.neighbours.write();
                     let n_to_add = PointWithOrder::<T>::new(&Arc::clone(&new_point), q.dist_to_ref);
-                    // must be sure that we add a point at the correct level
+                    // must be sure that we add a point at the correct level. See the comment to search_layer!
+                    // this ensures that reverse updating do not add problems.
                     let l_n = n_to_add.point_ref.p_id.0 as usize;
                     let already = q_point_neighbours[l_n as usize].iter().position(|old| old.point_ref.p_id == new_point.p_id);
                     if already.is_some() {
@@ -878,11 +948,10 @@ impl <T:Clone+Send+Sync, D: Distance<T>+Send+Sync > Hnsw<T,D>  {
                         continue;
                     }
                     q_point_neighbours[l_n].push(Arc::new(n_to_add));
-                    let nbn_at_l = q_point_neighbours[l_n].len();
-                    
+                    let nbn_at_l = q_point_neighbours[l_n].len();                    
                     //
                     // if l < level, update upward chaining, insert does a sort! t_q has a neighbour not yet in global table of points!
-                    let threshold_shrinking;  // TO DO optimize threshold
+                    let threshold_shrinking;  // TODO optimize threshold
                     if l_n > 0 {
                         threshold_shrinking = self.max_nb_connection;
                     }
@@ -1315,6 +1384,57 @@ fn test_iter_point() {
     } // end while
     //
     assert_eq!(nb_dumped, nbcolumn);
-} // end of test_insert_iter_point
+} // end of test_iter_point
+
+#[test]
+fn test_iter_layerpoint() {
+    //
+    println!("\n\n test_iter_point");
+    //
+    let mut rng = rand::thread_rng();
+    let unif =  Uniform::<f32>::new(0.,1.);
+    let nbcolumn = 5000;
+    let nbrow = 10;
+    let mut xsi;
+    let mut data = Vec::with_capacity(nbcolumn);
+    for j in 0..nbcolumn {
+        data.push(Vec::with_capacity(nbrow));
+        for _ in 0..nbrow {
+            xsi = rng.sample(unif);
+            data[j].push(xsi);
+        }
+    } 
+    //
+    // check insertion
+    let ef_construct= 25;
+    let nb_connection = 10;
+    let start = ProcessTime::now();
+    let hns = Hnsw::<f32, dist::DistL1>::new(nb_connection, nbcolumn, 16, ef_construct, dist::DistL1{});
+    for i in 0..data.len() {
+        hns.insert((&data[i], i));
+    }
+    let cpu_time = start.elapsed();
+    println!(" test_insert_iter_point time inserting {:?}", cpu_time); 
+
+    hns.dump_layer_info();
+    // now check iteration
+    let layer_num = 0;
+    let nbpl = hns.get_point_indexation().get_layer_nb_point(layer_num);
+    let mut layer_iter = hns.get_point_indexation().get_layer_iterator(layer_num);
+    //
+    let mut nb_dumped = 0;
+    loop {
+        if let Some(_point) = layer_iter.next() {
+        //    println!("point : {:?}", _point.p_id);
+            nb_dumped += 1;
+        }
+        else {
+            break;
+        }
+    } // end while
+    println!("test_iter_layerpoint : nb point in layer {} , nb found {}", nbpl, nb_dumped);
+    //
+    assert_eq!(nb_dumped, nbpl);
+} // end of test_iter_layerpoint
 
 }  // end of module test
